<?php
/**
 * WP CLI commands for DMG Media
 *
 * @package DMG_WP_CLI
 */

/**
 * Plugin Name: DMG CLI
 * Plugin URI:  https://github.com/ANMedia/dmg-cli-wp-plugin
 * Description: WP CLI commands for DMG Media
 * Version:     1.6.0
 * Author:      markparolisi
 */

if ( ! class_exists( 'WPCOM_VIP_CLI_Command' ) ) {
	return;
}

/**
 * CLI commands for DMG Media
 */
class DMG_WP_CLI extends WPCOM_VIP_CLI_Command {


	/**
	 * Helper method to output a dry-run message
	 *
	 * @param boolean $dry_run Whether or not doing a dry run
	 *
	 * @return void
	 */
	private static function dry_run_message( $dry_run = false ) {
		$dry_run_message = ( $dry_run ) ? 'Running in dry-run mode.' : 'Running in production mode!';
		WP_CLI::line( $dry_run_message );
	}

	/**
	 * Utility to remove a specific capability
	 *
	 * ## OPTIONS
	 *
	 * cap: the capability name
	 *
	 * ## EXAMPLES
	 *
	 * wp dmg user-remove-cap --cap="cap_name" --dry-run=true
	 *
	 * @subcommand  user-remove-cap
	 * @synopsis [--cap] [--dry-run=<bool>]
	 */
	public function user_remove_cap( $args, $assoc_args ) {

		$defaults = [
			'dry-run' => true,
			'cap'     => '',
		];

		$assoc_args = array_merge( $defaults, $assoc_args );

		$dry_run = ( isset( $assoc_args['dry-run'] ) ) ? filter_var( $assoc_args['dry-run'], FILTER_VALIDATE_BOOLEAN ) : false;
		self::dry_run_message( $dry_run );

		if ( $assoc_args['cap'] !== '' ) {
			global $wp_roles;
			foreach ( array_keys( $wp_roles->roles ) as $role ) {
				WP_CLI::line( sprintf( 'Removing "%s" from "%s"', $assoc_args['cap'], $role ) );
				if ( ! $dry_run ) {
					$wp_roles->remove_cap( $role, $assoc_args['cap'] );
				}
			}
			WP_CLI::success( 'User Caps removed.' );
		} else {
			WP_CLI::error( 'Please specify the capability to remove. Usage: wp dmg remove-cap --cap=<cap>' );
		}
	}


	/**
	 * Set the role to subscriber if the user role is missing
	 *
	 * @subcommand user-fill-missing-role
	 *
	 * @synopsis [--dry-run]
	 *
	 * wp dmg user-fill-missing-role --dry-run=true
	 */
	public function user_fill_missing_role( $args, $assoc_args ) {

		$count = 0;

		$defaults = [
			'dry-run' => true,
		];

		$assoc_args = array_merge( $defaults, $assoc_args );

		$dry_run = isset( $assoc_args['dry-run'] ) ? filter_var( $assoc_args['dry-run'], FILTER_VALIDATE_BOOLEAN ) : true;

		self::dry_run_message( $dry_run );

		$query_args = [ 'orderby' => 'user_registered' ];
		$user_query = new \WP_User_Query( $query_args );

		if ( ! empty( $user_query->get_results() ) ) {
			$this->start_bulk_operation();

			foreach ( $user_query->get_results() as $user ) {

				if ( empty ( $user->roles ) ) {
					WP_CLI::log( "Missing role for user $user->ID - $user->user_login." );
					$count ++;
					if ( ! $dry_run ) {
						WP_CLI::runcommand( "user set-role $user->ID subscriber", [ 'exit_error' => false ] );
						sleep( 1 );
					}
				}
			}

			if ( ! $dry_run ) {
				WP_CLI::success( sprintf( '%d users successfully updated.', $count ) );
			} else {
				WP_CLI::success( sprintf( '%d users will be updated.', $count ) );
			}

			$this->end_bulk_operation();
		} else {
			WP_CLI::log( 'No users found' );
		}
	}

	/**
	 * Append a taxonomy to posts that meet separate taxonomy criteria
	 *
	 * ## OPTIONS
	 *
	 * <taxonomy>
	 * : The taxonomy of the term to append
	 *
	 * <term>
	 * : The term to append
	 *
	 * [--term-field=<term-field>]
	 * : The field the term is expressed as; default slug
	 *
	 * [--date-after=<date-after>]
	 * : Start date for querying posts
	 *
	 * [--date-before=<date-before>]
	 * : End date for querying posts
	 *
	 * [--<field>=<value>]
	 * : Args for searching existing posts
	 *
	 *   There are four possible parameters that can be passed:
	 *
	 *      1) --search-<taxonomy>=<term>: The taxonomy term to query by
	 *      2) --search-<taxonomy>-field=<field>: The field to query the search taxonomy term by; default slug
	 *      3) --search-<taxonomy>-include-children=<include>: Whether to include children in taxonomy term search;
	 *      4) --search-meta-<meta-key>=<meta-value>: The meta key/value pair to search;
	 *
	 * [--search-relationship=<search-relationship>]
	 * : Relationship of tax query; default OR
	 *
	 * [--search-string=<search-string>]
	 * : The text string to search post content for
	 *
	 * [--search-author-name=<search-string>]
	 * : The author to search;
	 *
	 * [--search-post-type=<post-type>]
	 * : The post type to search; default post
	 *
	 * [--search-post-status=<post-status>]
	 * : The post status to search; default publish
	 *
	 * [--case-sensitive=<case-sensitive>]
	 * : If search-string is passed, enforce case sensitivity
	 *
	 * [--ignore-shortcodes=<ignore-shortcodes>]
	 * : If search-string is passed, ignore shortcodes
	 *
	 * [--append=<append>]
	 * : If true, append terms; if false, replace existing terms; default true
	 *
	 * [--dry-run=<dry-run>]
	 * : Perform the command without updating the database
	 * 
	 * [--skip-posts-with-subterms=<skip-posts-with-subterms>]
	 * : For hierarchical taxonomies, skip term assignment when the post has a term that is not top-level
	 * 
	 * [--skip-posts-with-different-top-level-terms=<skip-posts-with-different-top-level-terms>]
	 * : For hierarchical taxonomies, skip term assignment when the post has an existing term in the taxonomy to be assigned, but the top-level term is different from that of the term to assign
	 * 
	 * [--backup-<taxonomy>-for-different-top-level-terms=<term>]
	 * : For hierarchical taxonomies, when term assignment would otherwise be skipped due to mismatching top-level terms, choose a backup term to be assigned instead
	 *
	 * ## EXAMPLES
	 *
	 * wp dmg auto-term post_tag pizza --search-post_tag=cheese --dry-run=true
	 *
	 * @subcommand auto-term
	 */
	public function auto_term( $args, $assoc_args ) {

		// Is this a dry run?
		$dry_run = ( isset( $assoc_args['dry-run'] ) ) ? filter_var( $assoc_args['dry-run'], FILTER_VALIDATE_BOOLEAN ) : false;
		self::dry_run_message( $dry_run );

		// Make sure the taxonomy is valid
		if ( empty( $args[0] ) ) {
			WP_CLI::error( 'No taxonomy specified.' );
		}

		$taxonomy = get_taxonomy( $args[0] );
		if ( ! is_a( $taxonomy, 'WP_Taxonomy' ) ) {
			WP_CLI::error( 'Invalid taxonomy.' );
		}

		// Make sure the term is valid
		if ( empty( $args[1] ) ) {
			WP_CLI::error( 'No term specified.' );
		}

		// Determine the term field
		$term_field = 'slug';
		if ( isset( $assoc_args['term-field'] ) ) {
			if ( in_array( $assoc_args['term-field'], array( 'term_id', 'name', 'slug ' ) ) ) {
				$term_field = $assoc_args['term-field'];
			} else {
				WP_CLI::error( 'Invalid term field passed: ' . $assoc_args['term-field'] );
			}
		}

		// Retrieve the term object
		$term = get_term_by( $term_field, $args[1], $taxonomy->name );
		if ( ! is_a( $term, 'WP_Term' ) ) {
			WP_CLI::error( 'Term with ' . $term_field . ' ' . $args[1] . ' in taxonomy ' . $taxonomy->name . ' not found.' );
		}

		// Defaults for other arguments
		$case_sensitive    = ( isset( $assoc_args['case-sensitive'] ) ) ? filter_var( $assoc_args['case-sensitive'], FILTER_VALIDATE_BOOLEAN ) : true;
		$ignore_shortcodes = ( isset( $assoc_args['ignore-shortcodes'] ) ) ? filter_var( $assoc_args['ignore-shortcodes'], FILTER_VALIDATE_BOOLEAN ) : true;
		$append            = ( isset( $assoc_args['append'] ) ) ? filter_var( $assoc_args['append'], FILTER_VALIDATE_BOOLEAN ) : true;

		// Bool for skipping posts with subterms
		$skip_posts_with_subterms = ( isset( $assoc_args['skip-posts-with-subterms'] ) ) ? filter_var( $assoc_args['skip-posts-with-subterms'], FILTER_VALIDATE_BOOLEAN ) : false;

		// Bool for skipping posts with subterms
		$skip_posts_with_different_top_level_terms = ( isset( $assoc_args['skip-posts-with-different-top-level-terms'] ) ) ? filter_var( $assoc_args['skip-posts-with-different-top-level-terms'], FILTER_VALIDATE_BOOLEAN ) : false;

		// Determine the post type
		$post_type = 'post';
		if ( isset( $assoc_args['search-post-type'] ) ) {
			if ( post_type_exists( $assoc_args['search-post-type'] ) ) {
				$post_type = $assoc_args['search-post-type'];
			} else {
				WP_CLI::error( 'Invalid post type specified: ' . $assoc_args['search-post-type'] );
			}
		}

		// Determine the post status
		$post_status = 'publish';
		if ( isset( $assoc_args['search-post-status'] ) ) {
			$post_statuses = get_post_statuses();
			$post_statuses['any'] = 'Any'; // Add any
			if ( isset( $post_statuses[ $assoc_args['search-post-status'] ] ) ) {
				$post_status = $assoc_args['search-post-status'];
			} else {
				WP_CLI::error( 'Invalid post status specified: ' . $assoc_args['search-post-status'] );
			}
		}

		// Baseline query args
		$query_args = array(
			'post_status' => $post_status,
			'post_type'   => $post_type,
		);

		// Search by author
		if ( isset( $assoc_args['search-author-name'] ) ) {
			$query_args['author_name'] = $assoc_args['search-author-name'];
		}

		// Search by meta key value
		foreach ( $assoc_args as $key => $value ) {
			if ( false === strpos( $key, 'search-meta-') ) {
				continue;
			}

			$meta_key = explode("search-meta-", $key );

			if ( isset( $meta_key[1] ) ) {
				$query_args['meta_query'] = array(
					array(
						'key' => $meta_key[1],
						'value' => $value,
						'compare' => '='
					)
				);
			}
		}

		$tax_queries = array();

		// Build a regex pattern
		$taxonomy_pattern = implode( '|', get_taxonomies() );

		// Look for search parameters
		foreach ( $assoc_args as $key => $value ) {

			// Match against the taxonomy term
			if ( ! preg_match( "/search-({$taxonomy_pattern})$/", $key, $matches ) ) {
				continue;
			}

			if ( empty( $matches[1] ) ) {
				continue;
			}

			$tax_query_taxonomy = $matches[1];

			// Determine the field to query by
			$tax_query_field = 'slug';
			if ( isset( $assoc_args["{$key}-field"] ) ) {
				if ( in_array( $assoc_args["{$key}-field"], array( 'term_id', 'name', 'slug' ), true ) ) {
					$tax_query_field = $assoc_args["{$key}-field"];
				} else {
					WP_CLI::error( "Invalid value passed for {$key}-field: " . $assoc_args["{$key}-field"] );
				}
			}

			// Do we want to include children for hierarchical terms?
			$tax_query_include_children = false;
			if ( isset( $assoc_args["{$key}-include-children"] ) ) {
				$tax_query_include_children = filter_var( $assoc_args["{$key}-include-children"], FILTER_VALIDATE_BOOLEAN );
			}

			// Add the tax query
			$tax_queries[] = array(
				'taxonomy'         => $tax_query_taxonomy,
				'field'            => $tax_query_field,
				'terms'            => $value,
				'include_children' => $tax_query_include_children,
			);
		}

		// Build tax_query array
		if ( ! empty( $tax_queries ) ) {

			$query_args['tax_query'] = array();

			// Specify search relationship if we have more than one tax query
			if ( count( $tax_queries ) > 1 ) {

				// Default to OR
				$search_relationship = 'OR';

				if ( isset( $assoc_args['search-relationship'] ) ) {
					if ( in_array( strtoupper( $assoc_args['search-relationship'] ), array( 'AND', 'OR' ), true ) ) {
						$search_relationship = strtoupper( $assoc_args['search-relationship'] );
					} else {
						WP_CLI::error( "Invalid value passed for search-relationship: " . $assoc_args['search-relationship'] );
					}
				}

				$query_args['tax_query']['relation'] = $search_relationship;
			}

			$query_args['tax_query'] = array_merge( $query_args['tax_query'], $tax_queries );
		}

		// Get date query parameters from CLI args
		$date_query = self::get_date_query( $assoc_args );
		if ( ! empty( $date_query ) ) {
			$query_args['date_query'] = $date_query;
		}

		// Specify a search string
		if ( ! empty( $assoc_args['search-string'] ) ) {
			$query_args['s'] = $assoc_args['search-string'];
		}

		$backup_term_for_different_top_level_terms = false;

		// Look for backup terms
		foreach ( $assoc_args as $key => $value ) {

			// Match against the taxonomy term
			if ( ! preg_match( "/backup-({$taxonomy_pattern})-for-different-top-level-terms$/", $key, $matches ) ) {
				continue;
			}

			if ( empty( $matches[1] ) ) {
				continue;
			}

			$maybe_term = get_term_by( 'slug', $value, $matches[1] );
			if ( is_a( $maybe_term, 'WP_Term' ) ) {
				$backup_term_for_different_top_level_terms = $maybe_term;
			}
		}

		// Args to pass to the iterator
		$iterate_args = array_merge(
			$assoc_args,
			array(
				'term'                                      => $term,
				'dry_run'                                   => $dry_run,
				'case_sensitive'                            => $case_sensitive,
				'ignore_shortcodes'                         => $ignore_shortcodes,
				'append'                                    => $append,
				'skip_posts_with_subterms'                  => $skip_posts_with_subterms,
				'skip_posts_with_different_top_level_terms' => $skip_posts_with_different_top_level_terms,
				'backup_term_for_different_top_level_terms' => $backup_term_for_different_top_level_terms
			)
		);

		// Iterate over posts with callback
		$this->iterate_posts( array( __CLASS__, 'auto_term_handler' ), $query_args, $iterate_args );
	}

	/**
	 * Callback for auto-term iterator
	 *
	 * @param WP_Post $post The post object
	 * @param array $args Callback arguments
	 */
	public static function auto_term_handler( $post, $args ) {

		WP_CLI::line( 'Checking post #' . $post->ID . ' -- ' . $post->post_title );

		// Make sure iterator has received a WP_Term object
		if ( empty( $args['term'] ) || ! is_a( $args['term'], 'WP_Term' ) ) {
			WP_CLI::error( 'The iterator needs a WP_Term object.' );
		}

		// Get the taxonomy we're assigning
		$taxonomy = get_taxonomy( $args['term']->taxonomy );

		// Special handling for hierarchical categories
		if ( $taxonomy->hierarchical === true ) {

			// Create a boolean to force a skip
			$skip = false;

			// Get all post terms
			$terms = wp_get_post_terms( $post->ID, $args['term']->taxonomy );

			// Skip post if already assigned a subterm in the target taxonomy
			if ( $args['skip_posts_with_subterms'] === true ) {
				foreach ( $terms as $term ) {
					if ( (int) $term->parent > 0 ) {
						WP_CLI::line( "Post #{$post->ID} returned in search results, but already contains subterm {$term->name} in the {$args['term']->taxonomy} taxonomy." );
						$skip = true;
					}
				}
			}

			// Skip post if the post has a term, and the top-level term is
			// different than that of the one being assigned
			if ( $args['skip_posts_with_different_top_level_terms'] === true ) {

				// Function to get top-level term
				$get_top_level_term = function( $term ) use ( &$get_top_level_term ) {

					// Parent is 0; this is a top-level term
					if ( $term->parent === 0 ) {
						return $term;
					}

					$parent = get_term_by( 'id', $term->parent, $term->taxonomy );

					// Couldn't retrieve a term for some reason; return ull
					if ( ! is_a( $parent, 'WP_Term' ) ) {
						return null;
					}

					return ( $parent->parent === 0 ) ? $parent : $get_top_level_term( $parent );
				};

				// Get the top-level term of the term to assign
				$top_level_term_term = $get_top_level_term( $args['term'] );

				$found_match_top_level_terms = false;

				// See if there are any matches for the top-level term
				if ( is_a( $top_level_term_term, 'WP_Term' ) ) {
					foreach ( $terms as $term ) {
						$top_level_term_post = $get_top_level_term( $term );
						if ( is_a( $top_level_term_post, 'WP_Term' )
							&& $top_level_term_post->term_id === $top_level_term_term->term_id
						) {
							$found_match_top_level_terms = true;
						}
					}
				}

				// Skip if no top-level term match found
				if ( ! $found_match_top_level_terms ) {

					if ( is_a( $args['backup_term_for_different_top_level_terms'], 'WP_Term' ) ) {

						$backup_term = $args['backup_term_for_different_top_level_terms'];

						WP_CLI::line( "Post #{$post->ID} returned in search results, but does not share a top-level term. Attempting to assign backup {$backup_term->taxonomy} {$backup_term->name} instead." );

						// Append the term
						if ( ! $args['dry_run'] ) {
							$result = wp_set_object_terms( $post->ID, $backup_term->term_id, $backup_term->taxonomy, true );
							if ( is_array( $result ) ) {
								$success = true;
							}
						} else {
							$success = true;
						}

						if ( $success ) {

							if ( ! $args['dry_run'] ) {
								WP_CLI::line( "Post #{$post->ID} was updated with backup {$backup_term->taxonomy} {$backup_term->name}." );
							} else {
								WP_CLI::line( "Post #{$post->ID} will be updated with backup {$backup_term->taxonomy} {$backup_term->name}." );
							}
						} else {

							WP_CLI::line( "Something went wrong appending the backup term {$backup_term->name} to post #{$post->ID}" );
						}

					} else {
						WP_CLI::line( "Post #{$post->ID} returned in search results, but does not share a top-level {$taxonomy->name} with {$args['term']->name}." );
					}

					$skip = true;
				}
			}

			// Exit early if hierarchical args determined we should skip
			if ( $skip ) {
				WP_CLI::line( "Skipping {$taxonomy->name} assignment for post #{$post->ID}." );
				return;
			}
		}

		// Do a secondary case sensitive text check for search terms
		if ( ! empty( $args['search-string'] ) ) {

			// Make sure search method obeys case sensitivity argument
			$search  = $args['search-string'];
			$content = $post->post_content;
			if ( $args['case_sensitive'] !== true ) {
				$search  = strtolower( $search );
				$content = strtolower( $content );
			}

			// If we are ignoring shortcodes, strip them out here
			if ( $args['ignore_shortcodes'] === true ) {
				$pattern = get_shortcode_regex();
				$content = preg_replace( '/' . $pattern . '/s', '', $content );
			}

			// Skip post due to case mismatch
			if ( strpos( $content, $search ) === false ) {
				WP_CLI::line( 'Post #' . $post->ID . ' returned in search results, but does not contain exact text match. Skipping...' );

				return;
			}
		}

		// Append the term
		if ( ! $args['dry_run'] ) {
			$result = wp_set_object_terms( $post->ID, $args['term']->term_id, $args['term']->taxonomy, $args['append'] );
			if ( is_array( $result ) ) {
				$success = true;
			}
		} else {
			$success = true;
		}

		if ( $success ) {

			if ( ! $args['dry_run'] ) {
				WP_CLI::line( 'Post #' . $post->ID . ' was updated with ' . $args['term']->taxonomy . ' ' . $args['term']->name . '.' );
			} else {
				WP_CLI::line( 'Post #' . $post->ID . ' will be updated with ' . $args['term']->taxonomy . ' ' . $args['term']->name . '.' );
			}
		} else {

			WP_CLI::line( 'Something went wrong appending terms to post #' . $post->ID );
		}
	}

	/**
	 * Append the parent category to posts
	 *
	 * ## OPTIONS
	 *
	 * [--category-slug=<category-slug>]
	 * : Category slug for querying posts
	 *
	 * [--date-after=<date-after>]
	 * : Start date for querying posts
	 *
	 * [--date-before=<date-before>]
	 * : End date for querying posts
	 *
	 * [--search-post-type=<post-type>]
	 * : The post type to search; default post
	 *
	 * [--dry-run=<dry-run>]
	 * : Perform the command without updating the database
	 *
	 * ## EXAMPLES
	 *
	 * wp dmg backfill-category-parent --category-slug=tv-soaps --date-after=2022-01-01 --date-before=2022-10-01 --dry-run=true
	 *
	 * @subcommand backfill-category-parent
	 */
	public function backfill_category_parent( $args, $assoc_args ) {

		$defaults = [
			'dry-run'       => true,
			'category-slug' => false,
		];

		$assoc_args = array_merge( $defaults, $assoc_args );

		// Is this a dry run?
		$dry_run = ( isset( $assoc_args['dry-run'] ) ) ? filter_var( $assoc_args['dry-run'], FILTER_VALIDATE_BOOLEAN ) : false;
		self::dry_run_message( $dry_run );

		// Determine the post type
		$post_type = 'post';
		if ( isset( $assoc_args['search-post-type'] ) ) {
			if ( post_type_exists( $assoc_args['search-post-type'] ) ) {
				$post_type = $assoc_args['search-post-type'];
			} else {
				WP_CLI::error( 'Invalid post type specified: ' . $assoc_args['search-post-type'] );
			}
		}

		// Baseline query args
		$query_args = array(
			'post_status' => 'publish',
			'post_type'   => $post_type,
			'fields'      => 'ids'
		);

		if ( $assoc_args['category-slug'] ) {
			$query_args['tax_query'] = [
				[
					'taxonomy' => 'category',
					'field'    => 'slug',
					'terms'    => $assoc_args['category-slug'],
				]
			];
		}


		// Get date query parameters from CLI args
		$date_query = self::get_date_query( $assoc_args );
		if ( ! empty( $date_query ) ) {
			$query_args['date_query'] = $date_query;
		}

		// Args to pass to the iterator
		$iterate_args = array_merge(
			$assoc_args,
			array(
				'dry_run' => $dry_run
			)
		);

		// Iterate over posts with callback
		$this->iterate_posts( array( __CLASS__, 'backfill_category_parent_handler' ), $query_args, $iterate_args );
	}

	/**
	 * Callback for backfill-category-parent iterator
	 *
	 * @param WP_Post $post The post object
	 * @param array $args Callback arguments
	 */
	public static function backfill_category_parent_handler( $post_id, $args ) {

		WP_CLI::line( 'Checking post #' . $post_id );

		$success           = false;
		$categories        = get_the_category( $post_id );
		$parent_categories = [];

		do {
			$current_category = array_shift( $categories );
			if ( isset( $current_category->category_parent ) && $current_category->category_parent !== 0 ) {
				$parent_categories = array_unique( array_merge( $parent_categories, [ get_category( $current_category->category_parent ) ] ), SORT_REGULAR );
				$categories        = array_unique( array_merge( $categories, [ get_category( $current_category->category_parent ) ] ), SORT_REGULAR );
			}
		} while ( count( $categories ) > 0 );

		foreach ( $parent_categories as $parent_category ) {
			if ( ! $args['dry_run'] ) {
				wp_set_post_categories( $post_id, $parent_category->term_id, true );
			}
			$success = true;
		}

		if ( $success ) {
			if ( ! $args['dry_run'] ) {
				WP_CLI::line( 'Post #' . $post_id . ' was updated with ' . implode( ', ', wp_list_pluck( $parent_categories, 'name' ) ) . '.' );
			} else {
				WP_CLI::line( 'Post #' . $post_id . ' will be updated with ' . implode( ', ', wp_list_pluck( $parent_categories, 'name' ) ) . '.' );
			}
		} else {
			WP_CLI::line( 'No parent category to append for post #' . $post_id );
		}
	}

	/**
	 * Helper to get date query param from array of CLI arguments
	 *
	 * This expects a "date-before" and/or a "date-after"
	 * array key to work properly
	 *
	 * @param array $args The CLI arguments
	 *
	 * @return array A date_query parameter ready for WP_Query
	 */
	public static function get_date_query( $args ) {

		// Add date query if set
		$date_query = [];

		// Set the date after parameter
		if ( isset( $args['date-after'] ) ) {
			$pieces = explode( '-', $args['date-after'] );
			if ( isset( $pieces[0] )
			     && isset( $pieces[1] )
			     && isset( $pieces[2] )
			     && wp_checkdate( (int) $pieces[1], (int) $pieces[2], (int) $pieces[0], $args['date-after'] )
			) {
				$date_query['after'] = [
					'year'  => (int) $pieces[0],
					'month' => (int) $pieces[1],
					'day'   => (int) $pieces[2],
				];
			} else {
				WP_CLI::error( 'Invalid "date after" supplied.' );
			}
		}

		// End date
		if ( isset( $args['date-before'] ) ) {
			$pieces = explode( '-', $args['date-before'] );
			if ( isset( $pieces[0] )
			     && isset( $pieces[1] )
			     && isset( $pieces[2] )
			     && wp_checkdate( (int) $pieces[1], (int) $pieces[2], (int) $pieces[0], $args['date-before'] )
			) {
				$date_query['before'] = [
					'year'  => (int) $pieces[0],
					'month' => (int) $pieces[1],
					'day'   => (int) $pieces[2],
				];
			} else {
				WP_CLI::error( 'Invalid "date before" supplied.' );
			}
		}

		return $date_query;
	}

	/**
	 * Wrapper around wp_query that performs pagination and cleanup. Correctly iterates over all
	 * posts, even if posts self-modify to be removed from the query.
	 *
	 * @param string function( \WP_Post $post ) callback that operates on a WP_Post
	 * @param array  WP_Query arguments to select posts
	 * @param array  Other arguments to pass to callback
	 * @param string Order of execution. 'ASC' for ascending, 'DESC' for descending.
	 * @param bool   Whether or not to silence logging of query results
	 */
	protected function iterate_posts( $callback, $query_args, $func_args = [], $order = 'ASC', $silent = false ) {

		$posts_per_page = 1; // Speed up first query to count total results
		$page           = 1;
		// This function tries to avoid the problem of missing posts if the CLI function performs an action that removes a post
		// from the query. It does it by running on the last object in the query and going in reverse order, effectively reversing
		// the ASC<->DESC order. We reverse the apparent order to the caller so that appears sensible still makes sense.
		if ( strcasecmp( $order, 'ASC' ) === 0 ) {
			$order = 'DESC';
		} else {
			$order = 'ASC';
		}

		// Pagination fails with 'p' query - change to use post_in if exists
		if ( isset( $query_args['p'] ) ) {
			if ( isset( $query_args['post__in'] ) ) {
				WP_CLI::warning( "'post__in' and 'p' query set - ignore 'post__in'" );
			}
			$query_args['post__in'] = [ $query_args['p'] ];
			unset( $query_args['p'] );
		}

		if ( ! isset( $query_args['posts_per_page'] ) ) {
			$query_args['posts_per_page'] = $posts_per_page;
		}

		if ( ! isset( $query_args['paged'] ) ) {
			$query_args['paged'] = $page;
		}

		if ( ! isset( $query_args['order'] ) ) {
			$query_args['order'] = $order;
		}

		// Run the query twice so that we can iterate in reverse order
		$query                        = new \WP_Query( $query_args );
		$total                        = $query->found_posts;
		$posts_per_page               = 100;
		$query_args['paged']          = floor( $total / $posts_per_page ) + 1;
		$query_args['posts_per_page'] = $posts_per_page;

		do {
			$query = new \WP_Query( $query_args );

			if ( ! $silent ) {
				WP_CLI::line( 'Query: ' . wp_json_encode( $query->query ) );
				WP_CLI::line( 'Found: ' . count( $query->posts ) );
			}

			// Iterate through found posts backwards so that we don't
			// Run into pagination issues if our CLI command removes
			// posts from the resulting query
			if ( $query->have_posts() ) {
				for ( $i = $query->post_count - 1; $i >= 0; $i -- ) {
					call_user_func( $callback, $query->posts[ $i ], $func_args );
				}
			}

			// Increase the page count
			$query_args['paged'] --;

			// Free up memory
			$this->vip_inmemory_cleanup();
		} while ( $query_args['paged'] >= 1 );

	}

	/**
	 * Revoke oAuth and session tokens for all WordPress Users
	 *
	 * @subcommand users-revoke-tokens
	 *
	 * @synopsis [--dry-run]
	 *
	 * wp dmg users-revoke-tokens --dry-run=true // Run this first to test the script. Permissions won't be revoked
	 */
	public function users_revoke_tokens( $args, $assoc_args ) {

		$count = 0;

		$defaults = [
			'dry-run' => true,
		];

		$assoc_args = array_merge( $defaults, $assoc_args );

		$dry_run = isset( $assoc_args['dry-run'] ) ? filter_var( $assoc_args['dry-run'], FILTER_VALIDATE_BOOLEAN ) : true;

		if ( $dry_run ) {
			WP_CLI::log( 'Doing a dry run. Tokens won\'t be revoked.' );
		}

		$query_args = [ 'orderby' => 'user_registered', 'order' => 'desc' ];

		$user_query = new \WP_User_Query( $query_args );

		if ( ! empty( $user_query->get_results() ) ) {
			$this->start_bulk_operation();

			foreach ( $user_query->get_results() as $user ) {

				WP_CLI::log( "Searching tokens for $user->ID - $user->user_login." );
				$user_meta = get_user_meta( $user->ID );
				foreach ( $user_meta as $key => $value ) {
					$found = 0;
					// Revoke access to all users except for administrators (just a security measure)
					if ( 'session_tokens' === $key && ! in_array( 'administrator', $user->roles ) ) {
						$found ++;
						WP_CLI::log( "Revoking $key" );
						if ( ! $dry_run ) {
							delete_user_meta( $user->ID, $key );
						}
					} else if ( strpos( $key, 'oauth2_access_' ) !== false ) { // Revoke all oauth2 tokens
						$found ++;
						WP_CLI::log( "Revoking $key" );
						if ( ! $dry_run ) {
							delete_user_meta( $user->ID, $key );
						}
					}
				}
				if ( ! $dry_run && $found ) {
					sleep( 1 );
				}

				$count ++;
			}

			if ( ! $dry_run ) {
				WP_CLI::success( sprintf( 'Tokens successfully revoked for %d users', $count ) );
			} else {
				WP_CLI::success( sprintf( 'Tokens will be revoked for %d users', $count ) );
			}

			$this->end_bulk_operation();
		} else {
			WP_CLI::log( 'No users found' );
		}
	}

	/**
	 * Get WordPress Users / Guest Author Match
	 *
	 * @subcommand users-guest-authors-match
	 *
	 * wp dmg users-guest-authors-match
	 */
	public function users_guest_authors_match() {

		global $coauthors_plus;

		$co_guest_authors = new \CoAuthors_Guest_Authors();

		$num   = 100;
		$paged = 1;

		WP_CLI::line( 'User ID,Login,Email,Registered,Role,WP count,GA count,GA ID,URL' );

		do {

			$query_args = [ 'orderby' => 'user_registered', 'order' => 'desc', 'number' => $num, 'paged' => $paged ];

			$user_query = new \WP_User_Query( $query_args );

			if ( ! empty( $user_query->get_results() ) ) {
				$this->start_bulk_operation();

				foreach ( $user_query->get_results() as $user ) {

					$wp_count = wpcom_vip_count_user_posts( $user->ID );

					$guest_author     = $co_guest_authors->get_guest_author_by( 'linked_account', $user->user_login );
					$guest_author_id  = $guest_author ? $guest_author->ID : '';
					$guest_author_url = $guest_author ? sprintf( '%spost.php?post=%d&action=edit', get_admin_url(), $guest_author->ID ) : '';
					$ga_count         = $guest_author ? $coauthors_plus->get_guest_author_post_count( $guest_author ) : '';


					WP_CLI::line( sprintf( '%d,%s,%s,%s,%s,%d,%d,%s,%s', $user->ID, $user->user_login, $user->user_email, $user->user_registered, implode( ',', $user->roles ), $wp_count, $ga_count, $guest_author_id, $guest_author_url ) );

				}

				$this->end_bulk_operation();
			}

			$paged ++;

		} while ( count( $user_query->get_results() ) );
	}

	/**
	 * Perform a search/replace on user email.
	 *
	 * ## DESCRIPTION
	 *
	 * Look through all users and do a search & replace on email address.
	 *
	 * ## OPTIONS
	 *
	 * <search>
	 * : The string to search for.
	 *
	 * <replace>
	 * : The string to replace with.
	 * 
	 * [--users]
	 *
	 * : Optionally set a comma-separated list of user logins and/or IDs to search and replace.
	 *
	 * [--dry-run]
	 *
	 * : By default the command will run in dry-run mode. Set to false to make edits.
	 *
	 * [--revert]
	 *
	 * : Previous values are stored upon editing for 7 days. Set this flag to revert within 7 days.
	 *
	 * ## EXAMPLES
	 *
	 * wp dmg users-search-replace "@inews.co.uk" "@theipaper.com"
	 *
	 * @subcommand users-search-replace
	 *
	 * @todo Add a 'field' parameter so allow search/replace on other fields.
	 */
	public function users_search_replace( $args, $assoc_args ) {

		$dry_run = ( isset( $assoc_args['dry-run'] ) ) ? filter_var( $assoc_args['dry-run'], FILTER_VALIDATE_BOOLEAN ) : true;
		self::dry_run_message( $dry_run );

		$users_input = ( isset( $assoc_args['users'] ) ) ? \explode( ',', $assoc_args['users'] ) : false;
		if ( $users_input ) {
			$users_input = array_map( function( $identifier) {
				if ( is_numeric( $identifier ) ) {
					return (int) $identifier;
				} else {
					// Strip quotes and whitespace.
					return trim( $identifier, '\'" ');
				}
			}, $users_input );
		}

		$revert = ( isset( $assoc_args['revert'] ) ) ? filter_var( $assoc_args['revert'], FILTER_VALIDATE_BOOLEAN ) : false;

		if ( ! isset( $args[0] ) || ! isset( $args[1] ) ) {
			\WP_CLI::error( 'You must specify search and replace strings.' );
		}

		$search  = ( isset( $args[0] ) ) ? \sanitize_text_field( $args[0] ) : '';
		$replace = ( isset( $args[1] ) ) ? \sanitize_text_field( $args[1] ) : '';
		if ( empty( $search ) ) {
			\WP_CLI::error( 'The search string is empty. This is unwise.' );
		}

		$user_search_args = [
			'search'         => '*' . $search . '*',
			'search_columns' => [ 'user_email' ],
			'order'          => 'ASC',
			'orderby'        => 'ID',
			'number'         => 1, // To get the user count first.
		];

		$users = new \WP_User_Query( $user_search_args );
		if ( empty( $users->get_results() ) ) {
			\WP_CLI::error( 'No users found.' );
		}

		$user_search_args['number']      = 100;
		$total                           = $users->get_total();
		$user_search_args['paged']       = floor( $total / $user_search_args['number'] ) + 1;
		$user_search_args['count_total'] = false;

		$updated = [];

		do {
			$users = new \WP_User_Query( $user_search_args );

			// Iterate through found posts backwards so that we don't
			// Run into pagination issues if our CLI command removes
			// posts from the resulting query
			if ( ! empty( $users->get_results() ) ) {

				// Temporarily disable email notifications to avoid spamming people.
				add_filter( 'send_email_change_email', '__return_false' );

				for ( $i = count( $users->get_results() ) - 1; $i >= 0; $i -- ) {
					$user_id    = $users->get_results()[ $i ]->ID;
					$user_login = $users->get_results()[ $i ]->user_login;
					
					// If a user list is provided, skip if the user isn't in the list.
					if ( $users_input && ! in_array( $user_id, $users_input, true ) && ! in_array( $user_login, $users_input, true ) ) {
						continue;
					}

					$user_email = $users->get_results()[ $i ]->data->user_email;
					$new_email  = ( $revert ) ?
						get_user_meta( $user_id, 'backup_user_email', true ) :
						str_ireplace( $search, $replace, $user_email );

					\WP_CLI::line( sprintf( 'Processing user %d', $user_id ) );
					\WP_CLI::line( sprintf( 'Changing %s to %s for user ID %d', $user_email, $new_email, $user_id ) );

					if ( $user_email === $new_email ) {
						\WP_CLI::warning( sprintf( 'Current and new email are the same for user %d. Skipping.', $user_id ) );
						continue;
					} else if ( empty( $user_email ) || empty( $new_email ) ) {
						\WP_CLI::warning( sprintf( 'Email is empty for user %d. Skipping.', $user_id ) );
						continue;
					}

					if ( ! $dry_run ) {

						// We only need to back up the email address if we're not reverting.
						if ( ! $revert ) {
							$backup = update_user_meta( $user_id, 'backup_user_email', $user_email );
							if ( false === $backup ) {
								\WP_CLI::warning( sprintf( 'Failed to back up existing email address for user %d, Skipping', $user_id ) );
								continue;
							}
						}

						// Now update the email address.
						$update = wp_update_user(
							[
								'ID'         => $user_id,
								'user_email' => $new_email,
							]
						);
						if ( is_wp_error( $update ) ) {
							\WP_CLI::warning( sprintf( 'Failed to update user with ID %d.', $user_id ) );
						} else if ( $revert )  {
							// On a revert let's delete the backup, providing the revert went okay.
							delete_user_meta( $user_id, 'backup_user_email' );
						}

					}

					$updated[] = [
						'user_id'    => $user_id,
						'prev_email' => $user_email,
						'new_email'  => $new_email,
						'updated'    => ( ! is_wp_error( $update ) && ! $dry_run ) ? 'yes' : 'no',
					];
				}
			}

			// Increase the page count
			$user_search_args['paged'] --;

			// Free up memory
			$this->vip_inmemory_cleanup();

		} while ( $user_search_args['paged'] >= 1 );

		// Print out the table showing what's updated.
		\WP_CLI\Utils\format_items( 'table', $updated, [ 'user_id', 'prev_email', 'new_email', 'updated' ] );
		\WP_CLI::success( sprintf( 'Updated %d users.', count( $updated ) ) );

	}

	/**
	 * Delete terms from a given taxonomy
	 *
	 *
	 * ## DESCRIPTION
	 *
	 * Loop through and delete all terms in a taxonomy
	 *
	 * ## OPTIONS
	 *
	 * dry-run: if false, migrate posts
	 * taxonomy: the taxonomy name
	 *
	 * ## EXAMPLES
	 *
	 * wp dmg delete-taxonomy-terms --taxonomy=metro_region --dry-run=true  // Perform a dry run
	 * wp dmg delete-taxonomy-terms --taxonomy=metro_region --dry-run=false // Production run
	 *
	 * @subcommand delete-taxonomy-terms
	 * @synopsis [--dry-run] [--taxonomy]
	 */
	public function delete_taxonomy_terms( $args, $assoc_args ) {

		$this->start_bulk_operation();

		// Taxonomy must be specified
		if ( empty( $assoc_args['taxonomy'] ) ) {
			WP_CLI::error( 'Taxonomy is not specified! Aborting...' );
		}

		// Taxonomy must be valid
		if ( ! taxonomy_exists( $assoc_args['taxonomy'] ) ) {
			WP_CLI::error( 'Invalid taxonomy! Aborting...' );
		}

		// Default arguments
		$defaults = [
			'dry-run' => true,
		];

		// Get arguments
		$assoc_args = array_merge( $defaults, $assoc_args );

		// Determine if this is a dry run
		$dry_run = ( isset( $assoc_args['dry-run'] ) ) ? filter_var( $assoc_args['dry-run'], FILTER_VALIDATE_BOOLEAN ) : false;

		if ( $dry_run ) {
			WP_CLI::log( 'Doing a dry run. No data will be updated.' );
		}

		$taxonomy = $assoc_args['taxonomy'];

		$terms = get_terms(
			[
				'taxonomy'   => $taxonomy,
				'hide_empty' => false,
			]
		);

		$counter = 0;

		// Loop through each term and delete
		foreach ( $terms as $term ) {

			if ( is_a( $term, 'WP_Term' ) ) {

				// Delete term or spoof deletion on dry run
				if ( ! $dry_run ) {
					$deleted_term = wp_delete_term( $term->term_id, $taxonomy );
					if ( $deleted_term === true ) {
						$counter ++;
					}
				} else {
					$deleted_term = true;
				}

				if ( is_wp_error( $deleted_term ) ) {
					WP_CLI::line( $deleted_term->get_error_message() );
				} elseif ( $deleted_term === true ) {
					WP_CLI::line(
						sprintf(
							'Deleted term %d %s from %s',
							$term->term_id,
							$term->name,
							$taxonomy
						)
					);
				}
			}
		}

		WP_CLI::line(
			sprintf(
				'Deleted %d terms from %s',
				$counter,
				$taxonomy
			)
		);

		$this->end_bulk_operation();

	}

	/**
	 * Resync Apple News articles.
	 *
	 * ## OPTIONS
	 *
	 * dry-run: if false, perform the update
	 * date-after: start date for posts to resync
	 * date-before: end date for posts to resync
	 * pending-only: only process articles that are stuck in pending
	 *
	 * ## EXAMPLES
	 *
	 * wp dmg resync_apple_news_articles --date-before=2018-11-30 --date-after=2018-10-01 --dry-run=false
	 * wp dmg resync_apple_news_articles --date-before=2018-11-30 --date-after=2018-10-01 --dry-run=false --pending-only
	 *
	 * @subcommand resync_apple_news_articles
	 * @synopsis [--dry-run] [--date-after=<date>] [--date-before=<date>] [--pending-only]
	 */
	public function resync_apple_news_articles( $args, $assoc_args ) {

		if ( ! class_exists( '\Apple_News' ) ) {
			WP_CLI::error( 'Apple News not found! Check to see it the plugin is active.' );
		}

		// Determine if this is a dry run
		$dry_run      = isset( $assoc_args['dry-run'] ) ? filter_var( $assoc_args['dry-run'], FILTER_VALIDATE_BOOLEAN ) : false;
		$pending_only = isset( $assoc_args['pending-only'] ) ? true : false;

		// Notice for if we're doing a dry run or not.
		if ( $dry_run ) {
			WP_CLI::line( 'Running in dry-run mode.' );
		} else {
			WP_CLI::line( 'Running in production mode!' );
		}

		$date_after  = ( isset( $assoc_args['date-after'] ) ) ? strtotime( $assoc_args['date-after'] ) : gmdate( 'U' );
		$date_before = ( isset( $assoc_args['date-before'] ) ) ? strtotime( $assoc_args['date-before'] ) : gmdate( 'U' );

		$meta_query = [
			[
				'key'     => 'apple_news_api_id',
				'compare' => 'EXISTS',
			],
		];

		if ( $pending_only ) {
			WP_CLI::line( 'Only processing articles "stuck" in pending' );
			$meta_query[] = [
				'key'     => 'apple_news_api_pending',
				'compare' => 'EXISTS',
			];
		}

		$this->start_bulk_operation();

		$paged = 1;
		$total = 0;
		$error = 0;

		do {

			// Query posts with the apple_news_api_id meta key
			$query = new \WP_Query(
				[
					'post_type'      => 'post',
					'post_status'    => 'publish',
					'posts_per_page' => 100,
					'paged'          => $paged,
					'fields'         => 'ids',
					'meta_query'     => $meta_query,
					'date_query'     => [
						[
							//'inclusive' => true,
							'after'  => [
								'year'  => gmdate( 'Y', $date_after ),
								'month' => gmdate( 'm', $date_after ),
								'day'   => gmdate( 'd', $date_after ),
							],
							'before' => [
								'year'  => gmdate( 'Y', $date_before ),
								'month' => gmdate( 'm', $date_before ),
								'day'   => gmdate( 'd', $date_before ),
							],
						],
					],
				]
			);

			foreach ( $query->posts as $post_id ) {
				$an_id = get_post_meta( $post_id, 'apple_news_api_id', true );
				WP_CLI::line( 'INFO: Attempting to process #' . $post_id . ' which has an apple news id of ' . $an_id );

				if ( ! $pending_only ) {
					$pending = get_post_meta( $post_id, 'apple_news_api_pending', true );
					if ( ! empty( $pending ) ) {
						WP_CLI::line( 'INFO: #' . $post_id . ' is currently stuck in pending. Trying to force sync by removing pending flag.' );
						if ( ! $dry_run ) {
							delete_post_meta( $post_id, 'apple_news_api_pending' );
						}
					}
				} else {
					if ( ! $dry_run ) {
						delete_post_meta( $post_id, 'apple_news_api_pending' );
					}
				}

				$admin_settings = new \Admin_Apple_Settings();
				$settings       = $admin_settings->fetch_settings();

				if ( ! $dry_run ) {

					//pretend that the article is out of sync to have the plugin process accordingly.
					add_filter( 'apple_news_is_post_in_sync', '__return_false' );
					$action = new \Apple_Actions\Index\Push( $settings, $post_id );
					try {
						WP_CLI::line( 'INFO: Performing sync of #' . $post_id );
						$action->perform( true );
					} catch ( \Apple_Actions\Action_Exception $e ) {
						WP_CLI::line( 'ERROR: Failed to process ' . $post_id . ' -> ' . $e->getMessage() );
						$error ++;
					}
					remove_filter( 'apple_news_is_post_in_sync', '__return_false' );
				} else {
					WP_CLI::line( 'INFO: Sync would be performed for #' . $post_id );
				}

				$total ++;

				sleep( 15 );
			}

			$paged ++;

			// Free up memory
			$this->vip_inmemory_cleanup();

		} while ( count( $query->posts ) );

		if ( ! $dry_run ) {
			WP_CLI::line( $total . ' articles were processed by Apple News.' );
			WP_CLI::line( $error . ' articles had errors and were not synced with Apple News.' );
		} else {
			WP_CLI::line( $total . ' articles would be sent to Apple News.' );
		}
	}


	/**
	 * Delete Apple News articles.
	 *
	 * ## OPTIONS
	 *
	 * dry-run: if false, perform the update
	 * post-ids: CSV of post_ids to process
	 *
	 * ## EXAMPLES
	 *
	 * wp dmg delete_apple_news_articles --post-ids=8215683,8215563,8215712 --dry-run=false
	 * wp dmg delete_apple_news_articles --post-ids=8215683,8215563,8215712 --dry-run=true
	 *
	 * @subcommand delete_apple_news_articles
	 * @synopsis [--dry-run] [--post-ids]
	 */
	public function delete_apple_news_articles( $args, $assoc_args ) {

		if ( ! class_exists( '\Apple_News' ) ) {
			WP_CLI::error( 'Apple News not found! Check to see if the plugin is active.' );
		}

		$post_ids = isset( $assoc_args['post-ids'] ) ? explode( ',', $assoc_args['post-ids'] ) : [];

		if ( ! $post_ids ) {
			WP_CLI::error( 'post-ids isn\'t set. It should be a CSV of post ids to process' );
		}

		// Determine if this is a dry run
		$dry_run = isset( $assoc_args['dry-run'] ) ? filter_var( $assoc_args['dry-run'], FILTER_VALIDATE_BOOLEAN ) : false;

		// Notice for if we're doing a dry run or not.
		if ( $dry_run ) {
			WP_CLI::line( 'Running in dry-run mode.' );
		} else {
			WP_CLI::line( 'Running in production mode!' );
		}

		$this->start_bulk_operation();

		$total = 0;
		$error = 0;

		foreach ( $post_ids as $post_id ) {
			$an_id = get_post_meta( $post_id, 'apple_news_api_id', true );

			//Skip post ids that don't have an apple news api ID
			if ( ! $an_id ) {
				WP_CLI::line( 'INFO: Cannot process #' . $post_id . ' as it doesn\'t have an Apple News ID' );
				continue;
			}

			WP_CLI::line( 'INFO: Attempting to process #' . $post_id . ' which has an apple news id of ' . $an_id );

			$admin_settings = new \Admin_Apple_Settings();
			$settings       = $admin_settings->fetch_settings();

			if ( ! $dry_run ) {
				$action = new \Apple_Actions\Index\Delete( $settings, $post_id );
				try {
					WP_CLI::line( 'INFO: Performing deletion of #' . $post_id . ' from Apple News' );
					$action->perform( true );
				} catch ( \Apple_Actions\Action_Exception $e ) {
					WP_CLI::line( 'ERROR: Failed to delete ' . $post_id . ' from Apple News -> ' . $e->getMessage() );
					$error ++;
				}
			} else {
				WP_CLI::line( 'INFO: Deletion from Apple News would be performed for #' . $post_id );
			}

			$total ++;

			sleep( 15 );

			$this->vip_inmemory_cleanup();
		}

		if ( ! $dry_run ) {
			WP_CLI::line( $total . ' articles were deleted from Apple News.' );
			WP_CLI::line( $error . ' articles had errors and were not deleted from Apple News.' );
		} else {
			WP_CLI::line( $total . ' articles would be deleted from Apple News.' );
		}
	}

	/**
	 * Get a list of guest authors who meet some requirements
	 *
	 * ## OPTIONS
	 *
	 * post-threshold: Minimum number of user's posts
	 * additional-fields: Comma separated list of additional fields to show in the audit e.g --additional-fields="twitter,author-job-title"
	 * order-by: Sort the retrieved records by parameter. it accepts the same parameter of WP_Query.
	 * order: Assign the ascending or descending order. e.g. ASC, DESC
	 * active-users: Filter by active users
	 * active-users-range: Time range do to define active user
	 * active-users-threshold: Minimum number of user's posts in the time range to define an active user e.g. --active-users=true --active-users-range="30 days" --active-users-threshold="2"
	 *
	 * ## EXAMPLES
	 *
	 * wp dmg guest-author-audit --post-threshold=30 --additional-fields="twitter,author-job-title" --order-by=title --order=ASC --active-users=true --active-users-range="30 days" --active-users-threshold="2"
	 *
	 * @synopsis [--post-threshold] [--additional-fields] [--active-users] [--active-users-range] [--active-users-threshold] [--order-by] [--order]
	 *
	 * @subcommand guest-author-audit
	 */
	public function guest_author_audit( $args, $assoc_args ) {
		global $coauthors_plus;

		$this->start_bulk_operation();

		$defaults = [
			'post-threshold'         => 0,   // minimum number of user's posts
			'additional-fields'      => '',
			'active-users'           => false,
			'active-users-range'     => '30 days',
			'active-users-threshold' => '2',
			'order-by'               => 'date',
			'order'                  => 'DESC'
		];

		$assoc_args = array_merge( $defaults, $assoc_args );

		$active_users = filter_var( $assoc_args['active-users'], FILTER_VALIDATE_BOOLEAN );

		$fields = [
			'ID',
			'Guest author',
			'Unique slug',
			'First name',
			'Last name',
			'Creation date',
			'Linked account',
			'Guest Author count',
			'Post count(WP_Query)',
			'Edit page',
			'Avatar',
			'Bio',
		];

		$additional_fields = explode( ',', $assoc_args['additional-fields'] );
		$fields            = array_merge( $fields, $additional_fields );

		$header = implode( ',', $fields );

		WP_CLI::line( $header );

		$paged = 1;
		do {
			$query_args = [
				'post_type'           => 'guest-author',
				'post_status'         => 'any',
				'ignore_sticky_posts' => true,
				'posts_per_page'      => 100,
				'paged'               => $paged,
				'orderby'             => $assoc_args['order-by'],
				'order'               => $assoc_args['order']
			];
			$query      = new \WP_Query( $query_args );
			foreach ( $query->posts as $author ) {

				if ( function_exists( 'get_coauthors' ) ) {
					$co_author = $coauthors_plus->get_coauthor_by( 'ID', $author->ID );
					$count     = $coauthors_plus->get_guest_author_post_count( $co_author );
				}

				if ( ! $co_author ) {
					continue;
				}

				if ( $count >= $assoc_args['post-threshold'] ) {

					$posts = new WP_Query(
						array(
							'author_name' => $author->post_name,
							'post_type'   => 'post',
							'fields'      => 'ids'
						)
					);

					if ( $active_users ) {
						$latest_posts = new WP_Query(
							array(
								'author_name' => $author->post_name,
								'post_type'   => 'post',
								'post_status' => 'publish',
								'date_query'  => array(
									array(
										'after'  => sprintf( '- %s', $assoc_args['active-users-range'] ),
										'column' => 'post_date',
									),
								),
								'fields'      => 'ids'
							)
						);

						if ( $latest_posts->found_posts < $assoc_args['active-users-threshold'] ) {
							continue;
						}
					}

					// Default values always available out of the box
					$values = [
						$author->ID,
						$co_author->display_name,
						$co_author->user_nicename,
						$co_author->first_name,
						$co_author->last_name,
						$author->post_date,
						$co_author->linked_account,
						$count,
						$posts->found_posts,
						get_admin_url() . "post.php?post=$author->ID&action=edit",
						get_post_meta( $author->ID, '_thumbnail_id', true ),
						str_replace( '"', '', $co_author->description ),
					];

					// Additional fields defined in the CLI --additional-fields param
					foreach ( $additional_fields as $additional_field ) {
						$value = isset( $co_author->{$additional_field} ) ? $co_author->{$additional_field} : '';
						array_push( $values, $value );
					}

					WP_CLI::line( vsprintf( implode( ',', array_fill( 0, count( $fields ), '"%s"' ) ), str_replace( '"', '', $values ) ) );
				}
			}
			$paged ++;
			// Free up memory
			$this->vip_inmemory_cleanup();
		} while ( count( $query->posts ) );

		$this->end_bulk_operation();
	}

	/**
	 * ## DESCRIPTION
	 *
	 * Query all terms in a taxonomy and retrieve post count by date intervals
	 *
	 * ## OPTIONS
	 *
	 * @synopsis [--taxonomy=<taxonomy>] [--intervals=<intervals>]
	 *
	 * @subcommand taxonomy-post-count
	 *
	 * ## EXAMPLES
	 *
	 *  wp dmg taxonomy-post-count --taxonomy=post_tag --intervals="1 year ago,2 years ago,5 years ago"
	 */
	public function taxonomy_post_count( $args, $assoc_args ) {

		// Default arguments
		$defaults = [
			'taxonomy'  => 'category',
			'intervals' => '3 months ago,12 months ago,5 years ago'
		];

		// Get arguments
		$assoc_args = array_merge( $defaults, $assoc_args );

		$taxonomy       = $assoc_args['taxonomy'];
		$page           = 0;
		$terms_per_page = 100;

		// First row of output will be headers
		$fields    = array(
			'ID',
			'Name',
			'Slug',
			'URL',
			'Count: Total',
		);
		$intervals = explode( ',', $assoc_args['intervals'] );

		// Add the defined intervals
		foreach ( $intervals as $interval ) {
			$additional_fields[] = sprintf( 'Last %s', trim( str_replace( 'ago', '', $interval ) ) );
		}
		$fields = array_merge( $fields, $additional_fields );

		$header = implode( ',', $fields );

		WP_CLI::line( $header );

		$this->start_bulk_operation();

		do {

			// Get all tax terms
			$terms = get_terms(
				array(
					'taxonomy'   => $taxonomy,
					'hide_empty' => false,
					'orderby'    => 'name',
					'number'     => $terms_per_page,
					'offset'     => $page * $terms_per_page,
				)
			);

			foreach ( $terms as $term ) {

				$query_args = array(
					'post_type'   => 'post',
					'post_status' => 'publish',
					'fields'      => 'ids',
					'tax_query'   => array(
						array(
							'taxonomy' => $taxonomy,
							'field'    => 'term_id',
							'terms'    => $term->term_id,
						)
					)
				);

				$values = array(
					$term->term_id,
					$term->name,
					$term->slug,
					get_term_link( $term ),
					$term->count,
				);

				// Set the query args by defined intervals
				foreach ( $intervals as $interval ) {
					$query_args['date_query'][0]['after'] = $interval;
					$query                                = new WP_Query( $query_args );
					array_push( $values, $query->found_posts );
				}

				WP_CLI::line( implode( ',', $values ) );

			}

			// Increase the page count
			$page ++;

			// Free up memory
			$this->vip_inmemory_cleanup();

		} while ( count( $terms ) );

		$this->end_bulk_operation();
	}

	/**
	 * Evaluate published posts within a specified time frame to evaluate specified block usage
	 *
	 * ## OPTIONS
	 *
	 * block-name: block name e.g metro-gutenberg/link-shortcode
	 * post-type: post type to query posts
	 * date-after: start date to query posts
	 * date-before: end date to query posts
	 * count-reusable-blocks: count the contents of reusable blocks
	 * count-reusable-blocks: count the contents of innerBlocks
	 *
	 * ## EXAMPLES
	 *
	 * wp dmg block-usage --block-name="metro-gutenberg/link-shortcode" --date-before="2022-02-28" --date-after="2022-02-01" --count-reusable-blocks=true --count-inner-blocks=true
	 *
	 * @subcommand block-usage
	 * @synopsis [--block-name=<block>] [--post-type=<post_type>] [--date-after=<date>] [--date-before=<date>] [--count-reusable-blocks=<bool>] [--count-reusable-blocks=<bool>]
	 */
	public function block_usage( $args, $assoc_args ) {

		// Early exit if no block name
		if( ! isset( $assoc_args['block-name']  ) ) {
			WP_CLI::error( 'No block name supplied' );
		}

		$this->start_bulk_operation();

		// Build date query parameters
		$date_query = array();

		// Set the date after parameter
		if ( isset( $assoc_args['date-after'] ) ) {
			$pieces = explode( '-', $assoc_args['date-after'] );
			if ( isset( $pieces[0] )
				&& isset( $pieces[1] )
				&& isset( $pieces[2] )
				&& wp_checkdate( (int) $pieces[1], (int) $pieces[2], (int) $pieces[0], $assoc_args['date-after'] )
			) {
				$date_query['after'] = array(
					'year'  => (int) $pieces[0],
					'month' => (int) $pieces[1],
					'day'   => (int) $pieces[2],
				);
			} else {
				WP_CLI::error( 'Invalid "date after" supplied.' );
			}
		}

		// End date
		if ( isset( $assoc_args['date-before'] ) ) {
			$pieces = explode( '-', $assoc_args['date-before'] );
			if ( isset( $pieces[0] )
				&& isset( $pieces[1] )
				&& isset( $pieces[2] )
				&& wp_checkdate( (int) $pieces[1], (int) $pieces[2], (int) $pieces[0], $assoc_args['date-before'] )
			) {
				$date_query['before'] = array(
					'year'  => (int) $pieces[0],
					'month' => (int) $pieces[1],
					'day'   => (int) $pieces[2],
				);
			} else {
				WP_CLI::error( 'Invalid "date before" supplied.' );
			}
		}

		// Baseline query args
		$query = [
			'post_status' => 'publish',
			'post_type'   => ( ! empty( $assoc_args['post-type'] ) ) ? $assoc_args['post-type'] : 'post',
		];

		if ( ! empty( $date_query ) ) {
			$date_query['inclusive'] = true;
			$query['date_query']     = $date_query;
		}

		$fields = [
			'post_id',
			'post_date',
			'has_blocks',
			'has_block_' . $assoc_args['block-name'],
			'block_count',
			'parent_category',
			'sub_category',
			'link'
		];

		$header = implode( ',', $fields );

		WP_CLI::line( $header );

		// Iterate over posts with callback
		$this->iterate_posts(
			[ __CLASS__, 'block_usage_handler' ],
			$query,
			[
				'fields'                => $fields,
				'block_name'            => $assoc_args['block-name'],
				'count_reusable_blocks' => ( isset( $assoc_args['count-reusable-blocks'] ) ) ? filter_var( $assoc_args['count-reusable-blocks'], FILTER_VALIDATE_BOOLEAN ) : false, // Default to false
				'count_inner_blocks'    => ( isset( $assoc_args['count-inner-blocks'] ) ) ? filter_var( $assoc_args['count-inner-blocks'], FILTER_VALIDATE_BOOLEAN ) : false, // Default to false
			],
			'ASC',
			true
		);

		$this->end_bulk_operation();
	}

	/**
	 * Handler to check for misleading posts
	 *
	 * @param WP_Post $post The post object
	 * @param array   $args Function arguments
	 */
	public static function block_usage_handler( $post, $args ) {
		$block_name = $args['block_name'];
		$has_block  = has_block( $block_name, $post );
		$terms = self::get_cat_and_subcat( $post->ID );
		$count = 0;

		$blocks = parse_blocks( $post->post_content );

		// Count instances of block
		if ( $has_block ) {
			$filtered = array_filter( $blocks, function( $block ) use ( $block_name ) {
				return $block['blockName'] === $block_name;
			} );

			$count = count( $filtered );
		}

		// Count inner blocks
		foreach ( $blocks as $block ) {
			// Optionally include inner blocks in count
			if( $args['count_inner_blocks' ] ){
				$count += self::count_inner_blocks( $block, $block_name );
			}

			// Optionally include reusable blocks in count
			if ( $args['count_reusable_blocks' ] ) {

				if ( 'core/block' === $block['blockName'] && isset( $block['attrs']['ref'] ) ) {

					// Get corresponding post object
					$reusable_block = get_post( $block['attrs']['ref'] );

					if ( is_a( $reusable_block, 'WP_Post' ) ) {
						$reusable_block_blocks = parse_blocks( $reusable_block->post_content );

						$reusable_filtered = array_filter( $reusable_block_blocks, function( $block ) use ( $block_name ) {
							return $block['blockName'] === $block_name;
						} );

						$count += count( $reusable_filtered );

						foreach ( $reusable_block_blocks as $reusable_block_block ) {
							if( $args['count_inner_blocks' ] ){
								$count += self::count_inner_blocks( $reusable_block_block, $block_name );
							}
						}
					}
				}
			}
		}

		$data = [
			$post->ID,
			$post->post_date,
			has_blocks( $post ) ? 'true' : 'false',
			$has_block ? 'true' : 'false',
			$count,
			isset( $terms['category'] ) && is_a( $terms['category'], 'WP_Term' ) ? $terms['category']->slug : '',
			isset( $terms['subcategory'] ) && is_a( $terms['subcategory'], 'WP_Term' ) ?$terms['subcategory']->slug : '',
			get_permalink( $post->ID )
		];

		$line = implode( ',', $data );

		WP_CLI::line( $line );
	}

	/**
	 * Count instances of inner blocks
	 * 
	 * @param array $block       The block
	 * @param string $block_name The name of the block to look for
	 * @return int The number of occurences of the block
	 */
	public static function count_inner_blocks( $block, $block_name ) {
		$count = 0;

		if ( ! empty( $block['innerBlocks'] ) ) {
			foreach ( $block['innerBlocks'] as $inner_block ) {
				if ( $inner_block['blockName'] === $block_name ) {
					$count++;
				}

				if ( ! empty( $inner_block['innerBlocks'] ) ) {
					$count += self::count_inner_blocks( $inner_block, $block_name );
				}
			}
		}

		return $count;
	}

	/**
	 * Returns first category (and subcategory, if there is one)
	 *
	 * @param int $post_id
	 * @return array The category objects
	 */
	public static function get_cat_and_subcat( $post_id ) {
		$return     = array();
		$categories = get_the_category( $post_id );
		foreach ( $categories as $i => $c ) {
			if ( $c->category_parent == 0 ) {
				$return['category'] = $c;
				unset( $categories[ $i ] );
				foreach ( $categories as $s ) {
					if ( $s->category_parent == $return['category']->cat_ID ) {
						$return['subcategory'] = $s;
						break;
					}
				}

				return $return;
			}
		}

		if ( isset( $categories[0] ) && is_a( $categories[0], 'WP_Term' ) ) {
			// If we haven't found a category, grab a subcategory and infer the category
			$return['subcategory'] = $categories[0];
			$return['category']    = get_category( $categories[0]->category_parent );
		}

		return $return;

	}

		/**
	 * Evaluate published posts within a specified time frame to see term usage for a given taxonomy
	 *
	 * ## OPTIONS
	 *
	 * taxonomy: Taxono0my to evaluate
	 * post-type: post type to query posts
	 * date-after: start date to query posts
	 * date-before: end date to query posts
	 *
	 * ## EXAMPLES
	 *
	 * wp dmg term-usage --taxonomy="post_tag" --date-before="2022-02-28" --date-after="2022-02-01"
	 *
	 * @subcommand term-usage
	 * @synopsis [--taxonomy=<taxonomy>] [--post-type=<post_type>] [--date-after=<date>] [--date-before=<date>]
	 */
	public function term_usage( $args, $assoc_args ) {

		// Early exit if no taxonomy name
		if( ! isset( $assoc_args['taxonomy']  ) ) {
			WP_CLI::error( 'No taxonomy name supplied' );
		}

		$this->start_bulk_operation();

		// Build date query parameters
		$date_query = array();

		// Set the date after parameter
		if ( isset( $assoc_args['date-after'] ) ) {
			$pieces = explode( '-', $assoc_args['date-after'] );
			if ( isset( $pieces[0] )
				&& isset( $pieces[1] )
				&& isset( $pieces[2] )
				&& wp_checkdate( (int) $pieces[1], (int) $pieces[2], (int) $pieces[0], $assoc_args['date-after'] )
			) {
				$date_query['after'] = array(
					'year'  => (int) $pieces[0],
					'month' => (int) $pieces[1],
					'day'   => (int) $pieces[2],
				);
			} else {
				WP_CLI::error( 'Invalid "date after" supplied.' );
			}
		}

		// End date
		if ( isset( $assoc_args['date-before'] ) ) {
			$pieces = explode( '-', $assoc_args['date-before'] );
			if ( isset( $pieces[0] )
				&& isset( $pieces[1] )
				&& isset( $pieces[2] )
				&& wp_checkdate( (int) $pieces[1], (int) $pieces[2], (int) $pieces[0], $assoc_args['date-before'] )
			) {
				$date_query['before'] = array(
					'year'  => (int) $pieces[0],
					'month' => (int) $pieces[1],
					'day'   => (int) $pieces[2],
				);
			} else {
				WP_CLI::error( 'Invalid "date before" supplied.' );
			}
		}

		// Baseline query args
		$query = [
			'post_status' => 'publish',
			'post_type'   => ( ! empty( $assoc_args['post-type'] ) ) ? $assoc_args['post-type'] : 'post',
		];

		if ( ! empty( $date_query ) ) {
			$date_query['inclusive'] = true;
			$query['date_query']     = $date_query;
		}

		$fields = [
			'post_id',
			'post_date',
			'term_count',
			'term_slugs',
			'parent_category',
			'sub_category',
			'link'
		];

		$header = implode( ',', $fields );

		WP_CLI::line( $header );

		// Iterate over posts with callback
		$this->iterate_posts(
			[ __CLASS__, 'term_usage_handler' ],
			$query,
			[
				'fields'    => $fields,
				'taxonomy'  => $assoc_args['taxonomy'],
			],
			'ASC',
			true
		);

		$this->end_bulk_operation();
	}

	/**
	 * Handler to find term-usage 
	 *
	 * @param WP_Post $post The post object
	 * @param array   $args Function arguments
	 */
	public static function term_usage_handler( $post, $args ) {
		$taxonomy = $args['taxonomy'];
		$cats = self::get_cat_and_subcat( $post->ID );

		$terms = wp_get_post_terms( $post->ID, $taxonomy, [ 'fields' => 'slugs' ] );
		$count = count( $terms );
		
		$data = [
			$post->ID,
			$post->post_date,
			$count,
			implode( '|', $terms ),
			isset( $cats['category'] ) && is_a( $cats['category'], 'WP_Term' ) ? $cats['category']->slug : '',
			isset( $cats['subcategory'] ) && is_a( $cats['subcategory'], 'WP_Term' ) ? $cats['subcategory']->slug : '',
			get_permalink( $post->ID )
		];

		$line = implode( ',', $data );

		WP_CLI::line( $line );
	}

}

WP_CLI::add_command( 'dmg', 'DMG_WP_CLI' );
